#include "BC.cuh"

__device__ double atomicAddDouble(double* address, double val)
{
	unsigned long long int* address_as_ull =
		(unsigned long long int*)address;
	unsigned long long int old = *address_as_ull, assumed;
	do {
		assumed = old;
		old = atomicCAS(address_as_ull, assumed,
			__double_as_longlong(val +
			__longlong_as_double(assumed)));
	} while (assumed != old);
	return __longlong_as_double(old);
}
__device__  double atomicMinDouble(double *address, double val)
{
	unsigned long long ret = __double_as_longlong(*address);
	while (val < __longlong_as_double(ret))
	{
		unsigned long long old = ret;
		if ((ret = atomicCAS((unsigned long long *)address, old, __double_as_longlong(val))) == old)
			break;
	}
	return __longlong_as_double(ret);
}

__global__ void minEdge(int *nodes, Edge *edges, double * deltaNodes, int N) {
	int tid = blockIdx.x * blockDim.x + threadIdx.x;
	if (tid < N) {
		double minWeight = DBL_MAX;
		for(int i = nodes[tid]; i < nodes[tid+1]; i++) {
			if(edges[i].weight < minWeight) {
				minWeight = edges[i].weight;
			}
		}
		deltaNodes[tid] = minWeight;
	}
}

__global__ void devBC(int *nodes, Edge *edges, bool calNode, bool calEdge,
					  int *dev_U, int *dev_F, int *dev_ends, int *dev_S, double *dev_rst, long double *dev_sigma, 
					  double *dev_del, double *CB, double *ECB, double *deltaNodes, int *dev_lock, const int N, const int WARP_SIZE) {
	for (int s = blockIdx.x; s < N; s += gridDim.x) {
		int *U = dev_U + blockIdx.x * N;
		int *F = dev_F + blockIdx.x * N;
		double *rst = dev_rst + blockIdx.x * N;
		long double *sigma = dev_sigma + blockIdx.x * N;
		double *del = dev_del + blockIdx.x * N;
		int *ends = dev_ends + blockIdx.x * (N + 1);
		int *S = dev_S + blockIdx.x * N;
		int *lock = dev_lock + blockIdx.x * N;

		__shared__ int endsLen;
		__shared__ int SLen;
		__shared__ int FLen;
		__shared__ double delta;

		for (int i = threadIdx.x; i < N; i += blockDim.x) {
			U[i] = 1;
			rst[i] = DBL_MAX;
			sigma[i] = 0;
			del[i] = 0;
			ends[i] = 0;
			S[i] = 0;
			lock[i] = 0;
		}
		__syncthreads();
		if (threadIdx.x == 0) {
			rst[s] = 0.0;
			sigma[s] = 1;
			U[s] = 0;
			F[0] = s;
			FLen = 1;
			S[0] = s;
			SLen = 1;
			ends[0] = 0;
			ends[1] = 1;
			endsLen = 2;
			delta = 0.0;
		}
		__syncthreads();
		bool needlock = true;
		while (delta < DBL_MAX) {
			for (int j = threadIdx.x; j < FLen * WARP_SIZE; j += blockDim.x) {
				int i = F[j / WARP_SIZE];
				int num_adj = nodes[i+1] - nodes[i];
				Edge * adjs = & edges[nodes[i]];
				double rsti = rst[i];
				for(int k = j % WARP_SIZE; k < num_adj; k += WARP_SIZE) {
					Edge edge = adjs[k];
					int w = edge.v;
					double weight = edge.weight;
					double tmp = rsti + weight;
					needlock = true;
					while (needlock) {
						if (0 == atomicCAS(lock + w, 0, 1)) {
							if (tmp < rst[w]) {
								rst[w] = tmp;
								sigma[w] = 0;
							}
							if (tmp == rst[w]) {
								sigma[w] += sigma[i];
							}
							atomicExch(lock + w, 0);
							needlock = false;
						}
					}
				}
				__threadfence_block();
			}
			__syncthreads();
			if (threadIdx.x == 0) {
				delta = DBL_MAX;
			}
			__syncthreads();
			
			for (int i = threadIdx.x; i < N; i += blockDim.x) {
				if (U[i] == 1 && rst[i] < DBL_MAX) {
					double tmp = rst[i] + deltaNodes[i];
					atomicMinDouble(&delta, tmp);
				}
			}
			__syncthreads();
			
			if (threadIdx.x == 0) {
				FLen = 0;
			}
			__syncthreads();
			for (int i = threadIdx.x; i < N; i += blockDim.x) {
				if (U[i] == 1 && rst[i] < delta && rst[i] < DBL_MAX) {
					U[i] = 0; 
					int t = atomicAdd(&FLen, 1);
					F[t] = i;
				}
			}
			__syncthreads();
			if (threadIdx.x == 0) {
				if (FLen > 0) {
					ends[endsLen] = ends[endsLen - 1] + FLen;
					endsLen++;
					for(int i = 0; i < FLen; i++) {
						S[SLen + i] = F[i];
					}
					SLen += FLen;
				}
			}
			__syncthreads();
		}
		__syncthreads();
		__shared__ int depth, s_start, s_end;
		if (threadIdx.x == 0) {
			depth = endsLen - 1;
		}
		__syncthreads();
		
		while (depth > 0) {
			if (threadIdx.x == 0) {
				s_start = ends[depth - 1];
				s_end = ends[depth] - 1;
			}
			__syncthreads();
			int start = s_start;
			int end = s_end;
			for (int j = threadIdx.x; j < (end - start + 1) * WARP_SIZE; j += blockDim.x) {
				int i = start + j / WARP_SIZE;
				int w = S[i];
				int num_adj = nodes[w+1] - nodes[w];
				Edge * adjs = & edges[nodes[w]];
				long double sw = sigma[w];
				double rstw = rst[w];
				for(int k = j % WARP_SIZE; k < num_adj; k += WARP_SIZE) {
					Edge edge = adjs[k];
					int v = edge.v;
					double weight = edge.weight;
					double rstv = rst[v];
					if (!(rstv > rstw + weight) && !(rstv < rstw + weight)) {
						double c = (double)sw / sigma[v] * (1 + del[v]);
						if(calEdge) {
							atomicAddDouble(ECB + k, c);
						}
						// dsw += c;
						atomicAddDouble(del + w, c);
					}
				}
				__threadfence_block();
				// del[w] = dsw;
				if (j % WARP_SIZE == 0 && calNode && w != s) {
					atomicAddDouble(CB + w, del[w]);
				}
			}
			__syncthreads();
			if (threadIdx.x == 0){
				depth--;
			}
			__syncthreads();
		}
		__syncthreads();
	}
}
void BC(int N, int M, int *nodes, Edge *edges, bool calNode, bool calEdge, const char *bcpath, const char * ebcpath, const int WARP_SIZE) {
	int *dev_nodes;
	Edge *dev_edges;
	double *CB, *ECB;
	double *hostCB = new double[N];
	double *hostECB = new double[2 * M];
	cudaMalloc(&dev_nodes, (N+1) * sizeof(int));
	cudaMalloc(&dev_edges, 2 * M * sizeof(Edge));
	cudaMalloc(&CB, N * sizeof(double));
	cudaMalloc(&ECB, 2 * M * sizeof(double));
	cudaMemcpy(dev_nodes, nodes, (N+1) * sizeof(int), cudaMemcpyHostToDevice);
	cudaMemcpy(dev_edges, edges, 2 * M * sizeof(Edge), cudaMemcpyHostToDevice);
	memset(hostCB, 0, N * sizeof(double));
	cudaMemcpy(CB, hostCB, N * sizeof(double), cudaMemcpyHostToDevice);
	memset(hostECB, 0, 2 * M * sizeof(double));
	cudaMemcpy(ECB, hostECB, 2 * M * sizeof(double), cudaMemcpyHostToDevice);

	int blockCnt, threadsPerBlock;
	threadsPerBlock = 256;
	blockCnt = N / threadsPerBlock + (N % threadsPerBlock == 0 ? 0 : 1);
	double *deltaNodes;
	cudaMalloc(&deltaNodes, N * sizeof(double));
	minEdge << <blockCnt, threadsPerBlock >> >(dev_nodes, dev_edges, deltaNodes, N);
	threadsPerBlock = 256;
	blockCnt = 64;

	double *rst;
	long double *sigma;
	int *U, *F;
	int *lock;
	double *del;
	int *ends, *S;
	
	cudaMalloc(&rst, blockCnt * N * sizeof(double));
	cudaMalloc(&sigma, blockCnt * N * sizeof(long double));
	cudaMalloc(&del, blockCnt * N * sizeof(double));
	cudaMalloc(&U, blockCnt * N * sizeof(int));
	cudaMalloc(&F, blockCnt * N * sizeof(int));
	cudaMalloc(&lock, blockCnt * N * sizeof(int));
	cudaMalloc(&ends, blockCnt * (N + 1) * sizeof(int));
	cudaMalloc(&S, blockCnt * N * sizeof(int));

	devBC << <blockCnt, threadsPerBlock >> >(dev_nodes, 
		dev_edges, calNode, calEdge, U, F, ends, S, rst, sigma, del, CB, ECB, deltaNodes, lock, N, WARP_SIZE);
		
	cudaFree(rst);
	cudaFree(sigma);
	cudaFree(U);
	cudaFree(F);
	cudaFree(ends);
	cudaFree(S);
	cudaFree(del);
	cudaFree(lock);

	if(calNode) {
		cudaMemcpy(hostCB, CB, N * sizeof(double), cudaMemcpyDeviceToHost);
		FILE *fileBC = fopen(bcpath, "w");
		for (int i = 0; i < N; i++) {
			//hostCB[i] /= 2;
			fprintf(fileBC, "%d %lf\n", i, hostCB[i] / 2);
		}
		fclose(fileBC);
	}

	if(calEdge) {
		cudaMemcpy(hostECB, ECB, 2 * M * sizeof(double), cudaMemcpyDeviceToHost);
		FILE *fileEBC = fopen(ebcpath, "w");
		for(int w = 0; w < N; w++) {
			for(int i = nodes[w]; i < nodes[w+1]; i++) {
				Edge edge = edges[i];
				int v = edge.v;
				fprintf(fileEBC, "%d %d %lf\n", w, v, hostECB[i]);
			}
		}
		fclose(fileEBC);
	}
	cudaFree(dev_nodes);
	cudaFree(dev_edges);
	cudaFree(deltaNodes);
	cudaFree(CB);
	cudaFree(ECB);
	delete[] hostCB;
	delete[] hostECB;
}
